@@ -73,9 +73,8 @@ class _Collection(object):
             self.crt = int(time.mktime(d.timetuple()))
         self.sched = Scheduler(self)
         if not self.conf.get("newBury", False):
-            mod = self.db.mod
-            self.sched.unburyCards()
-            self.db.mod = mod
+            self.conf['newBury'] = True
+            self.setMod()
 
     def name(self):
         n = os.path.splitext(os.path.basename(self.path))[0]
@@ -148,14 +147,10 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
     def close(self, save=True):
         "Disconnect from DB."
         if self.db:
-            if not self.conf.get("newBury", False):
-                mod = self.db.mod
-                self.sched.unburyCards()
-                self.db.mod = mod
             if save:
                 self.save()
             else:
-                self.rollback()
+                self.db.rollback()
             if not self.server:
                 self.db.execute("pragma journal_mode = delete")
             self.db.close()
@@ -176,7 +171,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         self.load()
         self.lock()
 
-    def modSchema(self, check=True):
+    def modSchema(self, check):
         "Mark schema modified. Call this first so user can abort if necessary."
         if not self.schemaChanged():
             if check and not runFilter("modSchema", True):
@@ -202,7 +197,7 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         self.models.beforeUpload()
         self.tags.beforeUpload()
         self.decks.beforeUpload()
-        self.modSchema()
+        self.modSchema(check=False)
         self.ls = self.scm
         # ensure db is compacted before upload
         self.db.execute("vacuum")
@@ -309,12 +304,15 @@ crt=?, mod=?, scm=?, dty=?, usn=?, ls=?, conf=?""",
         snids = ids2str(nids)
         have = {}
         dids = {}
-        for id, nid, ord, did in self.db.execute(
-            "select id, nid, ord, did from cards where nid in "+snids):
+        for id, nid, ord, did, odid in self.db.execute(
+            "select id, nid, ord, did, odid from cards where nid in "+snids):
             # existing cards
             if nid not in have:
                 have[nid] = {}
             have[nid][ord] = id
+            # if in a filtered deck, add new cards to original deck
+            if odid != 0:
+                did = odid
             # and their dids
             if nid in dids:
                 if dids[nid] and dids[nid] != did:
@@ -384,7 +382,11 @@ insert into cards values (?,?,?,?,?,?,0,0,?,0,0,0,0,0,0,0,0,"")""",
         card = anki.cards.Card(self)
         card.nid = note.id
         card.ord = template['ord']
-        card.did = template['did'] or note.model()['did']
+        # Use template did (deck override) if valid, otherwise model did
+        if template['did'] and unicode(template['did']) in self.decks.decks:
+            card.did = template['did']
+        else:
+            card.did = note.model()['did']
         # if invalid did, use default instead
         deck = self.decks.get(card.did)
         if deck['dyn']:
@@ -503,6 +505,7 @@ where c.nid = n.id and c.id in %s group by nid""" % ids2str(cids)):
         fields['Tags'] = data[5].strip()
         fields['Type'] = model['name']
         fields['Deck'] = self.decks.name(data[3])
+        fields['Subdeck'] = fields['Deck'].split('::')[-1]
         if model['type'] == MODEL_STD:
             template = model['tmpls'][data[4]]
         else:
@@ -515,13 +518,11 @@ where c.nid = n.id and c.id in %s group by nid""" % ids2str(cids)):
         afmt = afmt or template['afmt']
         for (type, format) in (("q", qfmt), ("a", afmt)):
             if type == "q":
-                format = format.replace("{{cloze:", "{{cq:%d:" % (
-                    data[4]+1))
+                format = re.sub("{{(?!type:)(.*?)cloze:", r"{{\1cq-%d:" % (data[4]+1), format)
                 format = format.replace("<%cloze:", "<%%cq:%d:" % (
                     data[4]+1))
             else:
-                format = format.replace("{{cloze:", "{{ca:%d:" % (
-                    data[4]+1))
+                format = re.sub("{{(.*?)cloze:", r"{{\1ca-%d:" % (data[4]+1), format)
                 format = format.replace("<%cloze:", "<%%ca:%d:" % (
                     data[4]+1))
                 fields['FrontSide'] = stripSounds(d['q'])
@@ -613,13 +614,19 @@ where c.nid == f.id
             if self._undo[0] == 1:
                 old = self._undo[2]
             self.clearUndo()
-        self._undo = [1, _("Review"), old + [copy.copy(card)]]
+        wasLeech = card.note().hasTag("leech") or False
+        self._undo = [1, _("Review"), old + [copy.copy(card)], wasLeech]
 
     def _undoReview(self):
         data = self._undo[2]
+        wasLeech = self._undo[3]
         c = data.pop()
         if not data:
             self.clearUndo()
+        # remove leech tag if it didn't have it before
+        if not wasLeech and c.note().hasTag("leech"):
+            c.note().delTag("leech")
+            c.note().flush()
         # write old data
         c.flush()
         # and delete revlog entry
@@ -696,6 +703,11 @@ select id from notes where mid not in """ + ids2str(self.models.ids()))
             self.remNotes(ids)
         # for each model
         for m in self.models.all():
+            for t in m['tmpls']:
+                if t['did'] == "None":
+                    t['did'] = None
+                    problems.append(_("Fixed AnkiDroid deck override bug."))
+                    self.models.save(m)
             if m['type'] == MODEL_STD:
                 # model with missing req specification
                 if 'req' not in m:
@@ -753,6 +765,17 @@ select id from cards where odue > 0 and (type=1 or queue=2) and not odid""")
                          "Fixed %d cards with invalid properties.", cnt) % cnt)
             self.db.execute("update cards set odue=0 where id in "+
                 ids2str(ids))
+        # cards with odid set when not in a dyn deck
+        dids = [id for id in self.decks.allIds() if not self.decks.isDyn(id)]
+        ids = self.db.list("""
+select id from cards where odid > 0 and did in %s""" % ids2str(dids))
+        if ids:
+            cnt = len(ids)
+            problems.append(
+                ngettext("Fixed %d card with invalid properties.",
+                         "Fixed %d cards with invalid properties.", cnt) % cnt)
+            self.db.execute("update cards set odid=0, odue=0 where id in "+
+                ids2str(ids))
         # tags
         self.tags.registerNotes()
         # field cache
