@@ -3,19 +3,18 @@
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
 import re
+import traceback
 import urllib
 import unicodedata
 import sys
 import zipfile
 from cStringIO import StringIO
 
-import send2trash
 from anki.utils import checksum, isWin, isMac, json
 from anki.db import DB
 from anki.consts import *
 from anki.latex import mungeQA
 
-
 class MediaManager(object):
 
     soundRegexps = ["(?i)(\[sound:(?P<fname>[^]]+)\])"]
@@ -54,12 +53,54 @@ class MediaManager(object):
     def connect(self):
         if self.col.server:
             return
-        path = self.dir()+".db"
+        path = self.dir()+".db2"
         create = not os.path.exists(path)
         os.chdir(self._dir)
         self.db = DB(path)
         if create:
             self._initDB()
+        self.maybeUpgrade()
+
+    def _initDB(self):
+        self.db.executescript("""
+create table media (
+ fname text not null primary key,
+ csum text,           -- null indicates deleted file
+ mtime int not null,  -- zero if deleted
+ dirty int not null
+);
+
+create index idx_media_dirty on media (dirty);
+
+create table meta (dirMod int, lastUsn int); insert into meta values (0, 0);
+""")
+
+    def maybeUpgrade(self):
+        oldpath = self.dir()+".db"
+        if os.path.exists(oldpath):
+            self.db.execute('attach "../collection.media.db" as old')
+            try:
+                self.db.execute("""
+    insert into media
+     select m.fname, csum, mod, ifnull((select 1 from log l2 where l2.fname=m.fname), 0) as dirty
+     from old.media m
+     left outer join old.log l using (fname)
+     union
+     select fname, null, 0, 1 from old.log where type=1;""")
+                self.db.execute("delete from meta")
+                self.db.execute("""
+    insert into meta select dirMod, usn from old.meta
+    """)
+                self.db.commit()
+            except Exception, e:
+                # if we couldn't import the old db for some reason, just start
+                # anew
+                self.col.log("failed to import old media db:"+traceback.format_exc())
+            self.db.execute("detach old")
+            npath = "../collection.media.db.old"
+            if os.path.exists(npath):
+                os.unlink(npath)
+            os.rename("../collection.media.db", npath)
 
     def close(self):
         if self.col.server:
@@ -180,14 +221,18 @@ class MediaManager(object):
             txt = re.sub(reg, "", txt)
         return txt
 
-    def escapeImages(self, string):
+    def escapeImages(self, string, unescape=False):
+        if unescape:
+            fn = urllib.unquote
+        else:
+            fn = urllib.quote
         def repl(match):
             tag = match.group(0)
             fname = match.group("fname")
             if re.match("(https?|ftp)://", fname):
                 return tag
             return tag.replace(
-                fname, urllib.quote(fname.encode("utf-8")))
+                fname, unicode(fn(fname.encode("utf-8")), "utf8"))
         for reg in self.imgRegexps:
             string = re.sub(reg, repl, string)
         return string
@@ -268,79 +313,10 @@ class MediaManager(object):
     def have(self, fname):
         return os.path.exists(os.path.join(self.dir(), fname))
 
-    # Media syncing - changes and removal
-    ##########################################################################
-
-    def hasChanged(self):
-        return self.db.scalar("select 1 from log limit 1")
-
-    def removed(self):
-        return self.db.list("select * from log where type = ?", MEDIA_REM)
-
-    def syncRemove(self, fnames):
-        # remove provided deletions
-        for f in fnames:
-            if os.path.exists(f):
-                send2trash.send2trash(f)
-            self.db.execute("delete from log where fname = ?", f)
-            self.db.execute("delete from media where fname = ?", f)
-        # and all locally-logged deletions, as server has acked them
-        self.db.execute("delete from log where type = ?", MEDIA_REM)
-        self.db.commit()
-
-    # Media syncing - unbundling zip files from server
-    ##########################################################################
-
-    def syncAdd(self, zipData):
-        "Extract zip data; true if finished."
-        f = StringIO(zipData)
-        z = zipfile.ZipFile(f, "r")
-        finished = False
-        meta = None
-        media = []
-        # get meta info first
-        meta = json.loads(z.read("_meta"))
-        nextUsn = int(z.read("_usn"))
-        # then loop through all files
-        for i in z.infolist():
-            if i.filename == "_meta" or i.filename == "_usn":
-                # ignore previously-retrieved meta
-                continue
-            elif i.filename == "_finished":
-                # last zip in set
-                finished = True
-            else:
-                data = z.read(i)
-                csum = checksum(data)
-                name = meta[i.filename]
-                if not isinstance(name, unicode):
-                    name = unicode(name, "utf8")
-                # normalize name for platform
-                if isMac:
-                    name = unicodedata.normalize("NFD", name)
-                else:
-                    name = unicodedata.normalize("NFC", name)
-                # save file
-                open(name, "wb").write(data)
-                # update db
-                media.append((name, csum, self._mtime(name)))
-                # remove entries from local log
-                self.db.execute("delete from log where fname = ?", name)
-        # update media db and note new starting usn
-        if media:
-            self.db.executemany(
-                "insert or replace into media values (?,?,?)", media)
-        self.setUsn(nextUsn) # commits
-        # if we have finished adding, we need to record the new folder mtime
-        # so that we don't trigger a needless scan
-        if finished:
-            self.syncMod()
-        return finished
-
     # Illegal characters
     ##########################################################################
 
-    _illegalCharReg = re.compile(r'[][><:"/?*^\\|\0]')
+    _illegalCharReg = re.compile(r'[][><:"/?*^\\|\0\r\n]')
 
     def stripIllegal(self, str):
         return re.sub(self._illegalCharReg, "", str)
@@ -351,57 +327,16 @@ class MediaManager(object):
             return True
         return not not re.search(self._illegalCharReg, str)
 
-    # Media syncing - bundling zip files to send to server
-    ##########################################################################
-    # Because there's no standard filename encoding for zips, and because not
-    # all zip clients support retrieving mtime, we store the files as ascii
-    # and place a json file in the zip with the necessary information.
-
-    def zipAdded(self):
-        "Add files to a zip until over SYNC_ZIP_SIZE/COUNT. Return zip data."
-        f = StringIO()
-        z = zipfile.ZipFile(f, "w", compression=zipfile.ZIP_DEFLATED)
-        sz = 0
-        cnt = 0
-        files = {}
-        cur = self.db.execute(
-            "select fname from log where type = ?", MEDIA_ADD)
-        fnames = []
-        while 1:
-            fname = cur.fetchone()
-            if not fname:
-                # add a flag so the server knows it can clean up
-                z.writestr("_finished", "")
-                break
-            fname = fname[0]
-            # we add it as a one-element array simply to make
-            # the later forgetAdded() call easier
-            fnames.append([fname])
-            z.write(fname, str(cnt))
-            files[str(cnt)] = unicodedata.normalize("NFC", fname)
-            sz += os.path.getsize(fname)
-            if sz >= SYNC_ZIP_SIZE or cnt >= SYNC_ZIP_COUNT:
-                break
-            cnt += 1
-        z.writestr("_meta", json.dumps(files))
-        z.close()
-        return f.getvalue(), fnames
-
-    def forgetAdded(self, fnames):
-        if not fnames:
-            return
-        self.db.executemany("delete from log where fname = ?", fnames)
-        self.db.commit()
-
-    # Tracking changes (private)
+    # Tracking changes
     ##########################################################################
 
-    def _initDB(self):
-        self.db.executescript("""
-create table media (fname text primary key, csum text, mod int);
-create table meta (dirMod int, usn int); insert into meta values (0, 0);
-create table log (fname text primary key, type int);
-""")
+    def findChanges(self):
+        "Scan the media folder if it's changed, and note any changes."
+        if self._changed():
+            self._logChanges()
+
+    def haveDirty(self):
+        return self.db.scalar("select 1 from media where dirty=1 limit 1")
 
     def _mtime(self, path):
         return int(os.stat(path).st_mtime)
@@ -409,17 +344,6 @@ create table log (fname text primary key, type int);
     def _checksum(self, path):
         return checksum(open(path, "rb").read())
 
-    def usn(self):
-        return self.db.scalar("select usn from meta")
-
-    def setUsn(self, usn):
-        self.db.execute("update meta set usn = ?", usn)
-        self.db.commit()
-
-    def syncMod(self):
-        self.db.execute("update meta set dirMod = ?", self._mtime(self.dir()))
-        self.db.commit()
-
     def _changed(self):
         "Return dir mtime if it has changed since the last findChanges()"
         # doesn't track edits, but user can add or remove a file to update
@@ -429,38 +353,24 @@ create table log (fname text primary key, type int);
             return False
         return mtime
 
-    def findChanges(self):
-        "Scan the media folder if it's changed, and note any changes."
-        if self._changed():
-            self._logChanges()
-
     def _logChanges(self):
         (added, removed) = self._changes()
-        log = []
         media = []
-        mediaRem = []
         for f in added:
             mt = self._mtime(f)
-            media.append((f, self._checksum(f), mt))
-            log.append((f, MEDIA_ADD))
+            media.append((f, self._checksum(f), mt, 1))
         for f in removed:
-            mediaRem.append((f,))
-            log.append((f, MEDIA_REM))
+            media.append((f, None, 0, 1))
         # update media db
-        self.db.executemany("insert or replace into media values (?,?,?)",
+        self.db.executemany("insert or replace into media values (?,?,?,?)",
                             media)
-        if mediaRem:
-            self.db.executemany("delete from media where fname = ?",
-                                mediaRem)
         self.db.execute("update meta set dirMod = ?", self._mtime(self.dir()))
-        # and logs
-        self.db.executemany("insert or replace into log values (?,?)", log)
         self.db.commit()
 
     def _changes(self):
         self.cache = {}
         for (name, csum, mod) in self.db.execute(
-            "select * from media"):
+            "select fname, csum, mtime from media where csum is not null"):
             self.cache[name] = [csum, mod, False]
         added = []
         removed = []
@@ -475,9 +385,22 @@ create table log (fname text primary key, type int);
             if self.hasIllegal(f):
                 continue
             # empty files are invalid; clean them up and continue
-            if not os.path.getsize(f):
+            sz = os.path.getsize(f)
+            if not sz:
                 os.unlink(f)
                 continue
+            if sz > 100*1024*1024:
+                self.col.log("ignoring file over 100MB", f)
+                continue
+            # check encoding
+            if not isMac:
+                normf = unicodedata.normalize("NFC", f)
+                if f != normf:
+                    # wrong filename encoding which will cause sync errors
+                    if os.path.exists(normf):
+                        os.unlink(f)
+                    else:
+                        os.rename(f, normf)
             # newly added?
             if f not in self.cache:
                 added.append(f)
@@ -495,34 +418,112 @@ create table log (fname text primary key, type int);
                 removed.append(k)
         return added, removed
 
-    def sanityCheck(self):
-        assert not self.db.scalar("select count() from log")
-        cnt = self.db.scalar("select count() from media")
-        return cnt
+    # Syncing-related
+    ##########################################################################
+
+    def lastUsn(self):
+        return self.db.scalar("select lastUsn from meta")
+
+    def setLastUsn(self, usn):
+        self.db.execute("update meta set lastUsn = ?", usn)
+        self.db.commit()
+
+    def syncInfo(self, fname):
+        ret = self.db.first(
+            "select csum, dirty from media where fname=?", fname)
+        return ret or (None, 0)
+
+    def markClean(self, fnames):
+        for fname in fnames:
+            self.db.execute(
+                "update media set dirty=0 where fname=?", fname)
+
+    def syncDelete(self, fname):
+        if os.path.exists(fname):
+            os.unlink(fname)
+        self.db.execute("delete from media where fname=?", fname)
+
+    def mediaCount(self):
+        return self.db.scalar(
+            "select count() from media where csum is not null")
+
+    def dirtyCount(self):
+        return self.db.scalar(
+            "select count() from media where dirty=1")
 
     def forceResync(self):
         self.db.execute("delete from media")
-        self.db.execute("delete from log")
-        self.db.execute("update meta set usn = 0, dirMod = 0")
+        self.db.execute("update meta set lastUsn=0,dirMod=0")
+        self.db.execute("vacuum")
+        self.db.execute("analyze")
         self.db.commit()
 
-    def removeExisting(self, files):
-        "Remove files from list of files to sync, and return missing files."
-        need = []
-        remove = []
-        for f in files:
-            if isMac:
-                name = unicodedata.normalize("NFD", f)
+    # Media syncing: zips
+    ##########################################################################
+
+    def mediaChangesZip(self):
+        f = StringIO()
+        z = zipfile.ZipFile(f, "w", compression=zipfile.ZIP_DEFLATED)
+
+        fnames = []
+        # meta is list of (fname, zipname), where zipname of None
+        # is a deleted file
+        meta = []
+        sz = 0
+
+        for c, (fname, csum) in enumerate(self.db.execute(
+                        "select fname, csum from media where dirty=1"
+                        " limit %d"%SYNC_ZIP_COUNT)):
+
+            fnames.append(fname)
+            normname = unicodedata.normalize("NFC", fname)
+
+            if csum:
+                self.col.log("+media zip", fname)
+                z.write(fname, str(c))
+                meta.append((normname, str(c)))
+                sz += os.path.getsize(fname)
             else:
-                name = f
-            if self.db.scalar("select 1 from log where fname=?", name):
-                remove.append((name,))
+                self.col.log("-media zip", fname)
+                meta.append((normname, ""))
+
+            if sz >= SYNC_ZIP_SIZE:
+                break
+
+        z.writestr("_meta", json.dumps(meta))
+        z.close()
+        return f.getvalue(), fnames
+
+    def addFilesFromZip(self, zipData):
+        "Extract zip data; true if finished."
+        f = StringIO(zipData)
+        z = zipfile.ZipFile(f, "r")
+        media = []
+        # get meta info first
+        meta = json.loads(z.read("_meta"))
+        # then loop through all files
+        cnt = 0
+        for i in z.infolist():
+            if i.filename == "_meta":
+                # ignore previously-retrieved meta
+                continue
             else:
-                need.append(f)
-        self.db.executemany("delete from log where fname=?", remove)
-        self.db.commit()
-        # if we need all the server files, it's faster to pass None than
-        # the full list
-        if need and len(files) == len(need):
-            return None
-        return need
+                data = z.read(i)
+                csum = checksum(data)
+                name = meta[i.filename]
+                if not isinstance(name, unicode):
+                    name = unicode(name, "utf8")
+                # normalize name for platform
+                if isMac:
+                    name = unicodedata.normalize("NFD", name)
+                else:
+                    name = unicodedata.normalize("NFC", name)
+                # save file
+                open(name, "wb").write(data)
+                # update db
+                media.append((name, csum, self._mtime(name), 0))
+                cnt += 1
+        if media:
+            self.db.executemany(
+                "insert or replace into media values (?,?,?,?)", media)
+        return cnt
