@@ -2,17 +2,27 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-import urllib, os, sys, httplib2, gzip
+import urllib
+import sys
+import gzip
+import random
 from cStringIO import StringIO
+
+import httplib2
 from anki.db import DB
-from anki.utils import ids2str, intTime, json, isWin, isMac
+from anki.utils import ids2str, intTime, json, isWin, isMac, platDesc, checksum
 from anki.consts import *
 from hooks import runHook
+import anki
+from lang import ngettext
 
 # syncing vars
-HTTP_TIMEOUT = 30
+HTTP_TIMEOUT = 90
 HTTP_PROXY = None
 
+# badly named; means no retries
+httplib2.RETRIES = 1
+
 try:
     # httplib2 >=0.7.7
     _proxy_info_from_environment = httplib2.proxy_info_from_environment
@@ -37,7 +47,10 @@ def httpCon():
                 os.path.dirname(os.path.abspath(sys.argv[0])),
                 "../Resources/ankiweb.certs")
         else:
-            assert 0, "Your distro has not packaged Anki correctly."
+            certs = os.path.abspath(os.path.join(
+                os.path.dirname(certs), "..", "ankiweb.certs"))
+            if not os.path.exists(certs):
+                assert 0, "Your distro has not packaged Anki correctly."
     return httplib2.Http(
         timeout=HTTP_TIMEOUT, ca_certs=certs,
         proxy_info=HTTP_PROXY,
@@ -89,24 +102,54 @@ class Syncer(object):
         self.server = server
 
     def sync(self):
-        "Returns 'noChanges', 'fullSync', or 'success'."
+        "Returns 'noChanges', 'fullSync', 'success', etc"
+        self.syncMsg = ""
+        self.uname = ""
         # if the deck has any pending changes, flush them first and bump mod
         # time
         self.col.save()
         # step 1: login & metadata
         runHook("sync", "login")
-        ret = self.server.meta()
-        if not ret:
+        meta = self.server.meta()
+        self.col.log("rmeta", meta)
+        if not meta:
             return "badAuth"
-        self.rmod, rscm, self.maxUsn, rts, self.mediaUsn = ret
-        self.lmod, lscm, self.minUsn, lts, dummy = self.meta()
+        # server requested abort?
+        self.syncMsg = meta['msg']
+        if not meta['cont']:
+            return "serverAbort"
+        else:
+            # don't abort, but if 'msg' is not blank, gui should show 'msg'
+            # after sync finishes and wait for confirmation before hiding
+            pass
+        rscm = meta['scm']
+        rts = meta['ts']
+        self.rmod = meta['mod']
+        self.maxUsn = meta['usn']
+        # this is a temporary measure to address the problem of users
+        # forgetting which email address they've used - it will be removed
+        # when enough time has passed
+        self.uname = meta.get("uname", "")
+        meta = self.meta()
+        self.col.log("lmeta", meta)
+        self.lmod = meta['mod']
+        self.minUsn = meta['usn']
+        lscm = meta['scm']
+        lts = meta['ts']
         if abs(rts - lts) > 300:
+            self.col.log("clock off")
             return "clockOff"
         if self.lmod == self.rmod:
+            self.col.log("no changes")
             return "noChanges"
         elif lscm != rscm:
+            self.col.log("schema diff")
             return "fullSync"
         self.lnewer = self.lmod > self.rmod
+        # step 1.5: check collection is valid
+        if not self.col.basicCheck():
+            self.col.log("basic check")
+            return "basicCheckFailed"
         # step 2: deletions
         runHook("sync", "meta")
         lrem = self.removed()
@@ -122,6 +165,7 @@ class Syncer(object):
         while 1:
             runHook("sync", "stream")
             chunk = self.server.chunk()
+            self.col.log("server chunk", chunk)
             self.applyChunk(chunk=chunk)
             if chunk['done']:
                 break
@@ -130,6 +174,7 @@ class Syncer(object):
         while 1:
             runHook("sync", "stream")
             chunk = self.chunk()
+            self.col.log("client chunk", chunk)
             self.server.applyChunk(chunk=chunk)
             if chunk['done']:
                 break
@@ -140,9 +185,9 @@ class Syncer(object):
         if ret['status'] != "ok":
             # roll back and force full sync
             self.col.rollback()
-            self.col.modSchema()
+            self.col.modSchema(False)
             self.col.save()
-            raise Exception("collection sanity check failed")
+            return "sanityCheckFailed"
         # finalize
         runHook("sync", "finalize")
         mod = self.server.finish()
@@ -150,7 +195,15 @@ class Syncer(object):
         return "success"
 
     def meta(self):
-        return (self.col.mod, self.col.scm, self.col._usn, intTime(), None)
+        return dict(
+            mod=self.col.mod,
+            scm=self.col.scm,
+            usn=self.col._usn,
+            ts=intTime(),
+            musn=0,
+            msg="",
+            cont=True
+        )
 
     def changes(self):
         "Bundle up small objects."
@@ -182,16 +235,12 @@ class Syncer(object):
         self.prepareToChunk()
 
     def sanityCheck(self):
-        if self.col.db.scalar("""
-select count() from cards where nid not in (select id from notes)"""):
-            return "missing notes"
-        if self.col.db.scalar("""
-select count() from notes where id not in (select distinct nid from cards)"""):
-            return "missing cards"
+        if not self.col.basicCheck():
+            return "failed basic check"
         for t in "cards", "notes", "revlog", "graves":
             if self.col.db.scalar(
                 "select count() from %s where usn = -1" % t):
-                return "%t had usn = -1" % t
+                return "%s had usn = -1" % t
         for g in self.col.decks.all():
             if g['usn'] == -1:
                 return "deck had usn = -1"
@@ -274,7 +323,7 @@ from notes where %s""" % d)
 
     def chunk(self):
         buf = dict(done=False)
-        lim = 2500
+        lim = 250
         while self.tablesLeft and lim:
             curTable = self.tablesLeft[0]
             if not self.cursor:
@@ -442,6 +491,7 @@ from notes where %s""" % d)
         for r in data:
             if r[0] not in lmods or lmods[r[0]] < r[modIdx]:
                 update.append(r)
+        self.col.log(table, data)
         return update
 
     def mergeCards(self, cards):
@@ -451,7 +501,7 @@ from notes where %s""" % d)
             self.newerRows(cards, "cards", 4))
 
     def mergeNotes(self, notes):
-        rows = self.newerRows(notes, "notes", 4)
+        rows = self.newerRows(notes, "notes", 3)
         self.col.db.executemany(
             "insert or replace into notes values (?,?,?,?,?,?,?,?,?,?,?)",
             rows)
@@ -489,7 +539,9 @@ class HttpSyncer(object):
 
     def __init__(self, hkey=None, con=None):
         self.hkey = hkey
+        self.skey = checksum(str(random.random()))[:8]
         self.con = con or httpCon()
+        self.postVars = {}
 
     def assertOk(self, resp):
         if resp['status'] != '200':
@@ -501,17 +553,13 @@ class HttpSyncer(object):
     # costly. We could send it as a raw post, but more HTTP clients seem to
     # support file uploading, so this is the more compatible choice.
 
-    def req(self, method, fobj=None, comp=6,
-                 badAuthRaises=True, hkey=True):
+    def req(self, method, fobj=None, comp=6, badAuthRaises=False):
         BOUNDARY="Anki-sync-boundary"
         bdry = "--"+BOUNDARY
         buf = StringIO()
-        # compression flag and session key as post vars
-        vars = {}
-        vars['c'] = 1 if comp else 0
-        if hkey:
-            vars['k'] = self.hkey
-        for (key, value) in vars.items():
+        # post vars
+        self.postVars['c'] = 1 if comp else 0
+        for (key, value) in self.postVars.items():
             buf.write(bdry + "\r\n")
             buf.write(
                 'Content-Disposition: form-data; name="%s"\r\n\r\n%s\r\n' %
@@ -545,7 +593,7 @@ Content-Type: application/octet-stream\r\n\r\n""")
         body = buf.getvalue()
         buf.close()
         resp, cont = self.con.request(
-            SYNC_URL+method, "POST", headers=headers, body=body)
+            self.syncURL()+method, "POST", headers=headers, body=body)
         if not badAuthRaises:
             # return false if bad auth instead of raising
             if resp['status'] == '403':
@@ -561,11 +609,17 @@ class RemoteServer(HttpSyncer):
     def __init__(self, hkey):
         HttpSyncer.__init__(self, hkey)
 
+    def syncURL(self):
+        if os.getenv("ANKIDEV"):
+            return "https://l1sync.ankiweb.net/sync/"
+        return SYNC_BASE + "sync/"
+
     def hostKey(self, user, pw):
         "Returns hkey or none if user/pw incorrect."
+        self.postVars = dict()
         ret = self.req(
             "hostKey", StringIO(json.dumps(dict(u=user, p=pw))),
-            badAuthRaises=False, hkey=False)
+            badAuthRaises=False)
         if not ret:
             # invalid auth
             return
@@ -573,8 +627,13 @@ class RemoteServer(HttpSyncer):
         return self.hkey
 
     def meta(self):
+        self.postVars = dict(
+            k=self.hkey,
+            s=self.skey,
+        )
         ret = self.req(
-            "meta", StringIO(json.dumps(dict(v=SYNC_VER))),
+            "meta", StringIO(json.dumps(dict(
+                v=SYNC_VER, cv="ankidesktop,%s,%s"%(anki.version, platDesc())))),
             badAuthRaises=False)
         if not ret:
             # invalid auth
@@ -610,8 +669,17 @@ class FullSyncer(HttpSyncer):
 
     def __init__(self, col, hkey, con):
         HttpSyncer.__init__(self, hkey, con)
+        self.postVars = dict(
+            k=self.hkey,
+            v="ankidesktop,%s,%s"%(anki.version, platDesc()),
+        )
         self.col = col
 
+    def syncURL(self):
+        if os.getenv("ANKIDEV"):
+            return "https://l1.ankiweb.net/sync/"
+        return SYNC_BASE + "sync/"
+
     def download(self):
         runHook("sync", "download")
         self.col.close()
@@ -631,111 +699,215 @@ class FullSyncer(HttpSyncer):
         self.col = None
 
     def upload(self):
+        "True if upload successful."
         runHook("sync", "upload")
         # make sure it's ok before we try to upload
-        assert self.col.db.scalar("pragma integrity_check") == "ok"
+        if self.col.db.scalar("pragma integrity_check") != "ok":
+            return False
+        if not self.col.basicCheck():
+            return False
         # apply some adjustments, then upload
         self.col.beforeUpload()
-        assert self.req("upload", open(self.col.path, "rb")) == "OK"
+        if self.req("upload", open(self.col.path, "rb")) != "OK":
+            return False
+        return True
 
 # Media syncing
 ##########################################################################
+#
+# About conflicts:
+# - to minimize data loss, if both sides are marked for sending and one
+#   side has been deleted, favour the add
+# - if added/changed on both sides, favour the server version on the
+#   assumption other syncers are in sync with the server
+#
 
 class MediaSyncer(object):
 
     def __init__(self, col, server=None):
         self.col = col
         self.server = server
-        self.added = None
 
-    def sync(self, mediaUsn):
-        # step 1: check if there have been any changes
+    def sync(self):
+        # check if there have been any changes
         runHook("sync", "findMedia")
+        self.col.log("findChanges")
         self.col.media.findChanges()
-        lusn = self.col.media.usn()
-        if lusn == mediaUsn and not self.col.media.hasChanged():
+
+        # begin session and check if in sync
+        lastUsn = self.col.media.lastUsn()
+        ret = self.server.begin()
+        srvUsn = ret['usn']
+        if lastUsn == srvUsn and not self.col.media.haveDirty():
             return "noChanges"
-        # step 2: send/recv deletions
-        runHook("sync", "removeMedia")
-        lrem = self.removed()
-        rrem = self.server.remove(fnames=lrem, minUsn=lusn)
-        self.remove(rrem)
-        # step 3: stream files from server
-        runHook("sync", "server")
-        while 1:
-            runHook("sync", "streamMedia")
-            usn = self.col.media.usn()
-            zip = self.server.files(minUsn=usn)
-            if self.addFiles(zip=zip):
+
+        # loop through and process changes from server
+        self.col.log("last local usn is %s"%lastUsn)
+        self.downloadCount = 0
+        while True:
+            data = self.server.mediaChanges(lastUsn=lastUsn)
+
+            self.col.log("mediaChanges resp count %d"%len(data))
+            if not data:
                 break
-        # step 4: stream files to the server
-        runHook("sync", "client")
-        while 1:
-            runHook("sync", "streamMedia")
-            zip, fnames = self.files()
+
+            need = []
+            lastUsn = data[-1][1]
+            for fname, rusn, rsum in data:
+                lsum, ldirty = self.col.media.syncInfo(fname)
+                self.col.log(
+                    "check: lsum=%s rsum=%s ldirty=%d rusn=%d fname=%s"%(
+                        (lsum and lsum[0:4]),
+                        (rsum and rsum[0:4]),
+                        ldirty,
+                        rusn,
+                        fname))
+
+                if rsum:
+                    # added/changed remotely
+                    if not lsum or lsum != rsum:
+                        self.col.log("will fetch")
+                        need.append(fname)
+                    else:
+                        self.col.log("have same already")
+                    ldirty and self.col.media.markClean([fname])
+                elif lsum:
+                    # deleted remotely
+                    if not ldirty:
+                        self.col.log("delete local")
+                        self.col.media.syncDelete(fname)
+                    else:
+                        # conflict; local add overrides remote delete
+                        self.col.log("conflict; will send")
+                else:
+                    # deleted both sides
+                    self.col.log("both sides deleted")
+                    ldirty and self.col.media.markClean([fname])
+
+            self._downloadFiles(need)
+
+            self.col.log("update last usn to %d"%lastUsn)
+            self.col.media.setLastUsn(lastUsn) # commits
+
+        # at this point we're all up to date with the server's changes,
+        # and we need to send our own
+
+        updateConflict = False
+        toSend = self.col.media.dirtyCount()
+        while True:
+            zip, fnames = self.col.media.mediaChangesZip()
             if not fnames:
-                # finished
                 break
-            usn = self.server.addFiles(zip=zip)
-            # after server has replied, safe to remove from log
-            self.col.media.forgetAdded(fnames)
-            self.col.media.setUsn(usn)
-        # step 5: sanity check during beta testing
-        # NOTE: when removing this, need to move server tidyup
-        # back from sanity check to addFiles
-        s = self.server.mediaSanity()
-        c = self.mediaSanity()
-        if c != s:
-            raise Exception("""\
-Media sanity check failed. Please copy and paste the text below:\n%s\n%s""" %
-                            (c, s))
-        return "success"
 
-    def removed(self):
-        return self.col.media.removed()
+            runHook("syncMsg", ngettext(
+                "%d media change to upload", "%d media changes to upload", toSend)
+                    % toSend)
 
-    def remove(self, fnames, minUsn=None):
-        self.col.media.syncRemove(fnames)
-        if minUsn is not None:
-            # we're the server
-            return self.col.media.removed()
+            processedCnt, serverLastUsn = self.server.uploadChanges(zip)
+            self.col.media.markClean(fnames[0:processedCnt])
 
-    def files(self):
-        return self.col.media.zipAdded()
+            self.col.log("processed %d, serverUsn %d, clientUsn %d" % (
+                processedCnt, serverLastUsn, lastUsn
+            ))
 
-    def addFiles(self, zip):
-        "True if zip is the last in set. Server returns new usn instead."
-        return self.col.media.syncAdd(zip)
+            if serverLastUsn - processedCnt == lastUsn:
+                self.col.log("lastUsn in sync, updating local")
+                lastUsn = serverLastUsn
+                self.col.media.setLastUsn(serverLastUsn) # commits
+            else:
+                self.col.log("concurrent update, skipping usn update")
+                # commit for markClean
+                self.col.media.db.commit()
+                updateConflict = True
 
-    def mediaSanity(self):
-        return self.col.media.sanityCheck()
+            toSend -= processedCnt
+
+        if updateConflict:
+            self.col.log("restart sync due to concurrent update")
+            return self.sync()
+
+        lcnt = self.col.media.mediaCount()
+        ret = self.server.mediaSanity(local=lcnt)
+        if ret == "OK":
+            return "OK"
+        else:
+            self.col.media.forceResync()
+            return ret
+
+    def _downloadFiles(self, fnames):
+        self.col.log("%d files to fetch"%len(fnames))
+        while fnames:
+            top = fnames[0:SYNC_ZIP_COUNT]
+            self.col.log("fetch %s"%top)
+            zipData = self.server.downloadFiles(files=top)
+            cnt = self.col.media.addFilesFromZip(zipData)
+            self.downloadCount += cnt
+            self.col.log("received %d files"%cnt)
+            fnames = fnames[cnt:]
+
+            n = self.downloadCount
+            runHook("syncMsg", ngettext(
+                "%d media file downloaded", "%d media files downloaded", n)
+                    % n)
 
 # Remote media syncing
 ##########################################################################
 
 class RemoteMediaServer(HttpSyncer):
 
-    def __init__(self, hkey, con):
+    def __init__(self, col, hkey, con):
+        self.col = col
         HttpSyncer.__init__(self, hkey, con)
 
-    def remove(self, **kw):
-        return json.loads(
-            self.req("remove", StringIO(json.dumps(kw))))
+    def syncURL(self):
+        if os.getenv("ANKIDEV"):
+            return "https://l1.ankiweb.net/msync/"
+        return SYNC_MEDIA_BASE
 
-    def files(self, **kw):
-        return self.req("files", StringIO(json.dumps(kw)))
+    def begin(self):
+        self.postVars = dict(
+            k=self.hkey,
+            v="ankidesktop,%s,%s"%(anki.version, platDesc())
+        )
+        ret = self._dataOnly(json.loads(self.req(
+            "begin", StringIO(json.dumps(dict())))))
+        self.skey = ret['sk']
+        return ret
 
-    def addFiles(self, zip):
+    # args: lastUsn
+    def mediaChanges(self, **kw):
+        self.postVars = dict(
+            sk=self.skey,
+        )
+        resp = json.loads(
+            self.req("mediaChanges", StringIO(json.dumps(kw))))
+        return self._dataOnly(resp)
+
+    # args: files
+    def downloadFiles(self, **kw):
+        return self.req("downloadFiles", StringIO(json.dumps(kw)))
+
+    def uploadChanges(self, zip):
         # no compression, as we compress the zip file instead
-        return json.loads(
-            self.req("addFiles", StringIO(zip), comp=0))
+        return self._dataOnly(json.loads(
+            self.req("uploadChanges", StringIO(zip), comp=0)))
 
-    def mediaSanity(self):
-        return json.loads(
-            self.req("mediaSanity"))
+    # args: local
+    def mediaSanity(self, **kw):
+        return self._dataOnly(json.loads(
+            self.req("mediaSanity", StringIO(json.dumps(kw)))))
+
+    def _dataOnly(self, resp):
+        if resp['err']:
+            self.col.log("error returned:%s"%resp['err'])
+            raise Exception("SyncError:%s"%resp['err'])
+        return resp['data']
 
     # only for unit tests
-    def mediatest(self, n):
-        return json.loads(
-            self.req("mediatest", StringIO(
-                json.dumps(dict(n=n)))))
+    def mediatest(self, cmd):
+        self.postVars = dict(
+            k=self.hkey,
+        )
+        return self._dataOnly(json.loads(
+            self.req("newMediaTest", StringIO(
+                json.dumps(dict(cmd=cmd))))))
