@@ -7,7 +7,8 @@ from anki.consts import NEW_CARDS_RANDOM
 from aqt.qt import *
 import aqt
 from aqt.utils import showInfo, showWarning, openHelp, getOnlyText, askUser, \
-    tooltip
+    tooltip, saveGeom, restoreGeom, downArrow
+
 
 class DeckConf(QDialog):
     def __init__(self, mw, deck):
@@ -27,15 +28,16 @@ class DeckConf(QDialog):
                      SIGNAL("helpRequested()"),
                      lambda: openHelp("deckoptions"))
         self.connect(self.form.confOpts, SIGNAL("clicked()"), self.confOpts)
-        self.form.confOpts.setText(u"▾")
+        self.form.confOpts.setText(downArrow())
         self.connect(self.form.buttonBox.button(QDialogButtonBox.RestoreDefaults),
                      SIGNAL("clicked()"),
                      self.onRestore)
         self.setWindowTitle(_("Options for %s") % self.deck['name'])
         # qt doesn't size properly with altered fonts otherwise
+        restoreGeom(self, "deckconf", adjustSize=True)
         self.show()
-        self.adjustSize()
         self.exec_()
+        saveGeom(self, "deckconf")
 
     def setupCombos(self):
         import anki.consts as cs
