@@ -11,11 +11,6 @@ from anki.utils import isMac
 class EditCurrent(QDialog):
 
     def __init__(self, mw):
-        if isMac:
-            # use a separate window on os x so we can a clean menu
-            QDialog.__init__(self, None, Qt.Window)
-        else:
-            QDialog.__init__(self, mw)
         QDialog.__init__(self, None, Qt.Window)
         self.mw = mw
         self.form = aqt.forms.editcurrent.Ui_Dialog()
@@ -26,6 +21,8 @@ class EditCurrent(QDialog):
         self.connect(self,
                      SIGNAL("rejected()"),
                      self.onSave)
+        self.form.buttonBox.button(QDialogButtonBox.Close).setShortcut(
+                QKeySequence("Ctrl+Return"))
         self.editor = aqt.editor.Editor(self.mw, self.form.fieldsArea, self)
         self.editor.setNote(self.mw.reviewer.card.note())
         restoreGeom(self, "editcurrent")
