@@ -11,6 +11,7 @@ import random
 import cPickle
 import locale
 import re
+import shutil
 
 from aqt.qt import *
 from anki.db import DB
@@ -69,6 +70,7 @@ class ProfileManager(object):
             self.base = os.path.abspath(base)
         else:
             self.base = self._defaultBase()
+            self.maybeMigrateFolder()
         self.ensureBaseExists()
         # load metadata
         self.firstRun = self._loadMeta()
@@ -95,6 +97,36 @@ issue, please see the documentation for information on running Anki from \
 a flash drive.""" % self.base)
             raise
 
+    # Folder migration
+    ######################################################################
+
+    def _oldFolderLocation(self):
+        if isMac:
+            return os.path.expanduser("~/Documents/Anki")
+        elif isWin:
+            loc = QDesktopServices.storageLocation(QDesktopServices.DocumentsLocation)
+            return os.path.join(loc, "Anki")
+        else:
+            p = os.path.expanduser("~/Anki")
+            if os.path.exists(p):
+                return p
+            else:
+                loc = QDesktopServices.storageLocation(QDesktopServices.DocumentsLocation)
+                if loc[:-1] == QDesktopServices.storageLocation(
+                        QDesktopServices.HomeLocation):
+                    # occasionally "documentsLocation" will return the home
+                    # folder because the Documents folder isn't configured
+                    # properly; fall back to an English path
+                    return os.path.expanduser("~/Documents/Anki")
+                else:
+                    return os.path.join(loc, "Anki")
+
+    def maybeMigrateFolder(self):
+        oldBase = self._oldFolderLocation()
+
+        if not os.path.exists(self.base) and os.path.exists(oldBase):
+            shutil.move(oldBase, self.base)
+
     # Profile load/save
     ######################################################################
 
@@ -211,28 +243,18 @@ and no other programs are accessing your profile folders, then try again."""))
 
     def _defaultBase(self):
         if isWin:
-            if False: #qtmajor >= 5:
-                loc = QStandardPaths.writeableLocation(QStandardPaths.DocumentsLocation)
-            else:
-                loc = QDesktopServices.storageLocation(QDesktopServices.DocumentsLocation)
-            return os.path.join(loc, "Anki")
+            from win32com.shell import shell, shellcon
+            loc = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, None, 0)
+            dir = os.path.join(loc, "Anki2")
+            return dir
         elif isMac:
-            return os.path.expanduser("~/Documents/Anki")
+            return os.path.expanduser("~/Library/Application Support/Anki2")
         else:
-            # use Documents/Anki on new installs, ~/Anki on existing ones
-            p = os.path.expanduser("~/Anki")
-            if os.path.exists(p):
-                return p
-            else:
-                loc = QDesktopServices.storageLocation(QDesktopServices.DocumentsLocation)
-                if loc[:-1] == QDesktopServices.storageLocation(
-                        QDesktopServices.HomeLocation):
-                    # occasionally "documentsLocation" will return the home
-                    # folder because the Documents folder isn't configured
-                    # properly; fall back to an English path
-                    return os.path.expanduser("~/Documents/Anki")
-                else:
-                    return os.path.join(loc, "Anki")
+            dataDir = os.environ.get(
+                "XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
+            if not os.path.exists(dataDir):
+                os.makedirs(dataDir)
+            return os.path.join(dataDir, "Anki2")
 
     def _loadMeta(self):
         path = os.path.join(self.base, "prefs.db")
