@@ -2,17 +2,24 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
-import    difflib, re, cgi
+from __future__ import division
+import difflib
+import re
+import cgi
 import unicodedata as ucd
 import HTMLParser
+
+from anki.lang import _, ngettext
 from aqt.qt import *
 from anki.utils import  stripHTML, isMac, json
 from anki.hooks import addHook, runHook
 from anki.sound import playFromText, clearAudioQueue, play
-from aqt.utils import mungeQA, getBase, openLink, tooltip
+from aqt.utils import mungeQA, getBase, openLink, tooltip, askUserDialog, \
+    downArrow
 from aqt.sound import getAudio
 import aqt
 
+
 class Reviewer(object):
     "Manage reviews.  Maintains a separate state."
 
@@ -65,8 +72,14 @@ class Reviewer(object):
         elapsed = self.mw.col.timeboxReached()
         if elapsed:
             part1 = ngettext("%d card studied in", "%d cards studied in", elapsed[1]) % elapsed[1]
-            part2 = ngettext("%s minute.", "%s minutes.", elapsed[0]/60) % (elapsed[0]/60)
-            tooltip("%s %s" % (part1, part2), period=5000)
+            mins = int(round(elapsed[0]/60))
+            part2 = ngettext("%s minute.", "%s minutes.", mins) % mins
+            fin = _("Finish")
+            diag = askUserDialog("%s %s" % (part1, part2),
+                             [_("Continue"), fin])
+            diag.setIcon(QMessageBox.Information)
+            if diag.run() == fin:
+                return self.mw.moveToState("deckBrowser")
             self.mw.col.startTimebox()
         if self.cardQueue:
             # undone/edited cards to show
@@ -94,14 +107,19 @@ class Reviewer(object):
     # Audio
     ##########################################################################
 
-    def replayAudio(self):
+    def replayAudio(self, previewer=None):
+        if previewer:
+            state = previewer._previewState
+            c = previewer.card
+        else:
+            state = self.state
+            c = self.card
         clearAudioQueue()
-        c = self.card
-        if self.state == "question":
+        if state == "question":
             playFromText(c.q())
-        elif self.state == "answer":
+        elif state == "answer":
             txt = ""
-            if self._replayq(c):
+            if self._replayq(c, previewer):
                 txt = c.q()
             txt += c.a()
             playFromText(txt)
@@ -110,7 +128,7 @@ class Reviewer(object):
     ##########################################################################
 
     _revHtml = """
-<img src="qrc:/icons/rating.png" class=marked>
+<img src="qrc:/icons/rating.png" id=star class=marked>
 <div id=qa></div>
 <script>
 var ankiPlatform = "desktop";
@@ -122,13 +140,16 @@ function _updateQA (q, answerMode, klass) {
         typeans.focus();
     }
     if (answerMode) {
-        window.location = "#answer";
+        var e = $("#answer");
+        if (e[0]) { e[0].scrollIntoView(); }
     } else {
         window.scrollTo(0, 0);
     }
     if (klass) {
         document.body.className = klass;
     }
+    // don't allow drags of images, which cause them to be deleted
+    $("img").attr("draggable", false);
 };
 
 function _toggleStar (show) {
@@ -171,8 +192,7 @@ function _typeAnsPress() {
     ##########################################################################
 
     def _mungeQA(self, buf):
-        return self.mw.col.media.escapeImages(
-            self.typeAnsFilter(mungeQA(buf)))
+        return self.typeAnsFilter(mungeQA(self.mw.col, buf))
 
     def _showQuestion(self):
         self._reps += 1
@@ -182,10 +202,10 @@ function _typeAnsPress() {
         # grab the question and play audio
         if c.isEmpty():
             q = _("""\
-The front of this card is empty. Please run Tools>Maintenance>Empty Cards.""")
+The front of this card is empty. Please run Tools>Empty Cards.""")
         else:
             q = c.q()
-        if self._autoplay(c):
+        if self.autoplay(c):
             playFromText(q)
         # render & update bottom
         q = self._mungeQA(q)
@@ -200,13 +220,14 @@ The front of this card is empty. Please run Tools>Maintenance>Empty Cards.""")
         # user hook
         runHook('showQuestion')
 
-    def _autoplay(self, card):
+    def autoplay(self, card):
         return self.mw.col.decks.confForDid(
-            self.card.odid or self.card.did)['autoplay']
+            card.odid or card.did)['autoplay']
 
-    def _replayq(self, card):
-        return self.mw.col.decks.confForDid(
-            self.card.odid or self.card.did).get('replayq', True)
+    def _replayq(self, card, previewer=None):
+        s = previewer if previewer else self
+        return s.mw.col.decks.confForDid(
+            s.card.odid or s.card.did).get('replayq', True)
 
     def _toggleStar(self):
         self.web.eval("_toggleStar(%s);" % json.dumps(
@@ -223,7 +244,7 @@ The front of this card is empty. Please run Tools>Maintenance>Empty Cards.""")
         c = self.card
         a = c.a()
         # play audio?
-        if self._autoplay(c):
+        if self.autoplay(c):
             playFromText(a)
         # render and update bottom
         a = self._mungeQA(a)
@@ -276,8 +297,10 @@ The front of this card is empty. Please run Tools>Maintenance>Empty Cards.""")
             self.replayAudio()
         elif key == "*":
             self.onMark()
-        elif key == "-":
+        elif key == "=":
             self.onBuryNote()
+        elif key == "-":
+            self.onBuryCard()
         elif key == "!":
             self.onSuspend()
         elif key == "@":
@@ -315,8 +338,11 @@ The front of this card is empty. Please run Tools>Maintenance>Empty Cards.""")
 hr { background-color:#ccc; margin: 1em; }
 body { margin:1.5em; }
 img { max-width: 95%; max-height: 95%; }
-.marked { position:absolute; right: 7px; top: 7px; display: none; }
+.marked { position:fixed; right: 7px; top: 7px; display: none; }
 #typeans { width: 100%; }
+.typeGood { background: #0f0; }
+.typeBad { background: #f00; }
+.typeMissed { background: #ccc; }
 """
 
     def _styles(self):
@@ -325,8 +351,6 @@ img { max-width: 95%; max-height: 95%; }
     # Type in the answer
     ##########################################################################
 
-    failedCharColour = "#FF0000"
-    passedCharColour = "#00FF00"
     typeAnsPat = "\[\[type:(.+?)\]\]"
 
     def typeAnsFilter(self, buf):
@@ -362,7 +386,7 @@ img { max-width: 95%; max-height: 95%; }
             if self.typeCorrect is None:
                 if clozeIdx:
                     warn = _("""\
-Please run Tools>Maintenance>Empty Cards""")
+Please run Tools>Empty Cards""")
                 else:
                     warn = _("Type answer: unknown field %s") % fld
                 return re.sub(self.typeAnsPat, warn, buf)
@@ -377,28 +401,35 @@ Please run Tools>Maintenance>Empty Cards""")
 """ % (self.typeFont, self.typeSize), buf)
 
     def typeAnsAnswerFilter(self, buf):
-        if not self.typeCorrect:
-            return re.sub(self.typeAnsPat, "", buf)
         # tell webview to call us back with the input content
         self.web.eval("_getTypedText();")
+        if not self.typeCorrect:
+            return re.sub(self.typeAnsPat, "", buf)
+        origSize = len(buf)
+        buf = buf.replace("<hr id=answer>", "")
+        hadHR = len(buf) != origSize
         # munge correct value
         parser = HTMLParser.HTMLParser()
         cor = stripHTML(self.mw.col.media.strip(self.typeCorrect))
+        # ensure we don't chomp multiple whitespace
+        cor = cor.replace(" ", "&nbsp;")
         cor = parser.unescape(cor)
+        cor = cor.replace(u"\xa0", " ")
         given = self.typedAnswer
         # compare with typed answer
-        res = self.correct(cor, given)
-        if cor != given:
-            # Wrap the extra text in an id-ed span.
-            res += u"<span id=rightanswer><br> {0} <br> {1} </span>".format(
-                _(u"Correct answer was:"), cor)
+        res = self.correct(given, cor, showBad=False)
         # and update the type answer area
         def repl(match):
             # can't pass a string in directly, and can't use re.escape as it
             # escapes too much
-            return """
+            s = """
 <span style="font-family: '%s'; font-size: %spx">%s</span>""" % (
                 self.typeFont, self.typeSize, res)
+            if hadHR:
+                # a hack to ensure the q/a separator falls before the answer
+                # comparison when user is using {{FrontSide}}
+                s = "<hr id=answer>" + s
+            return s
         return re.sub(self.typeAnsPat, repl, buf)
 
     def _contentForCloze(self, txt, idx):
@@ -410,68 +441,74 @@ Please run Tools>Maintenance>Empty Cards""")
                 return txt.split("::")[0]
             return txt
         matches = [noHint(txt) for txt in matches]
-        if len(matches) > 1:
-            txt = ", ".join(matches)
-        else:
+        uniqMatches = set(matches)
+        if len(uniqMatches) == 1:
             txt = matches[0]
+        else:
+            txt = ", ".join(matches)
         return txt
 
-    # following type answer functions thanks to Bernhard
-    def calculateOkBadStyle(self):
-        "Precalculates styles for correct and incorrect part of answer"
-        st = "background: %s; color: #000;"
-        self.styleOk  = st % self.passedCharColour
-        self.styleBad = st % self.failedCharColour
+    def tokenizeComparison(self, given, correct):
+        # compare in NFC form so accents appear correct
+        given = ucd.normalize("NFC", given)
+        correct = ucd.normalize("NFC", correct)
+        try:
+            s = difflib.SequenceMatcher(None, given, correct, autojunk=False)
+        except:
+            # autojunk was added in python 2.7.1
+            s = difflib.SequenceMatcher(None, given, correct)
+        givenElems = []
+        correctElems = []
+        givenPoint = 0
+        correctPoint = 0
+        offby = 0
+        def logBad(old, new, str, array):
+            if old != new:
+                array.append((False, str[old:new]))
+        def logGood(start, cnt, str, array):
+            if cnt:
+                array.append((True, str[start:start+cnt]))
+        for x, y, cnt in s.get_matching_blocks():
+            # if anything was missed in correct, pad given
+            if cnt and y-offby > x:
+                givenElems.append((False, "-"*(y-x-offby)))
+                offby = y-x
+            # log any proceeding bad elems
+            logBad(givenPoint, x, given, givenElems)
+            logBad(correctPoint, y, correct, correctElems)
+            givenPoint = x+cnt
+            correctPoint = y+cnt
+            # log the match
+            logGood(x, cnt, given, givenElems)
+            logGood(y, cnt, correct, correctElems)
+        return givenElems, correctElems
 
-    def ok(self, a):
-        "returns given sring in style correct (green)"
-        if len(a) == 0:
-            return ""
-        return "<span style='%s'>%s</span>" % (self.styleOk, cgi.escape(a))
-
-    def bad(self, a):
-        "returns given sring in style incorrect (red)"
-        if len(a) == 0:
-            return ""
-        return "<span style='%s'>%s</span>" % (self.styleBad, cgi.escape(a))
-
-    def applyStyle(self, testChar, correct, wrong):
-        "Calculates answer fragment depending on testChar's unicode category"
-        ZERO_SIZE = 'Mn'
-        def head(a):
-            return a[:len(a) - 1]
-        def tail(a):
-            return a[len(a) - 1:]
-        if ucd.category(testChar) == ZERO_SIZE:
-            return self.ok(head(correct)) + self.bad(tail(correct) + wrong)
-        return self.ok(correct) + self.bad(wrong)
-
-    def correct(self, a, b):
+    def correct(self, given, correct, showBad=True):
         "Diff-corrects the typed-in answer."
-        if b == "":
-            return "";
-        self.calculateOkBadStyle()
-        ret = ""
-        lastEqual = ""
-        s = difflib.SequenceMatcher(None, b, a)
-        for tag, i1, i2, j1, j2 in s.get_opcodes():
-            if tag == "equal":
-                lastEqual = b[i1:i2]
-            elif tag == "replace":
-                ret += self.applyStyle(b[i1], lastEqual,
-                                 b[i1:i2] + ("-" * ((j2 - j1) - (i2 - i1))))
-                lastEqual = ""
-            elif tag == "delete":
-                ret += self.applyStyle(b[i1], lastEqual, b[i1:i2])
-                lastEqual = ""
-            elif tag == "insert":
-                if ucd.category(a[j1]) != 'Mn':
-                    dashNum = (j2 - j1)
+        givenElems, correctElems = self.tokenizeComparison(given, correct)
+        def good(s):
+            return "<span class=typeGood>"+cgi.escape(s)+"</span>"
+        def bad(s):
+            return "<span class=typeBad>"+cgi.escape(s)+"</span>"
+        def missed(s):
+            return "<span class=typeMissed>"+cgi.escape(s)+"</span>"
+        if given == correct:
+            res = good(given)
+        else:
+            res = ""
+            for ok, txt in givenElems:
+                if ok:
+                    res += good(txt)
                 else:
-                    dashNum = ((j2 - j1) - 1)
-                ret += self.applyStyle(a[j1], lastEqual, "-" * dashNum)
-                lastEqual = ""
-        return ret + self.ok(lastEqual)
+                    res += bad(txt)
+            res += "<br>&darr;<br>"
+            for ok, txt in correctElems:
+                if ok:
+                    res += good(txt)
+                else:
+                    res += missed(txt)
+        res = "<div><code id=typeans>" + res + "</code></div>"
+        return res
 
     # Bottom bar
     ##########################################################################
@@ -509,7 +546,7 @@ min-width: 60px; white-space: nowrap;
 </td>
 <td width=50 align=right valign=top class=stat><span id=time class=stattxt>
 </span><br>
-<button onclick="py.link('more');">%(more)s &#9662;</button>
+<button onclick="py.link('more');">%(more)s %(downArrow)s</button>
 </td>
 </tr>
 </table>
@@ -557,7 +594,9 @@ function showAnswer(txt) {
 </script>
 """ % dict(rem=self._remaining(), edit=_("Edit"),
            editkey=_("Shortcut key: %s") % "E",
-           more=_("More"), time=self.card.timeTaken()/1000)
+           more=_("More"),
+           downArrow=downArrow(),
+           time=self.card.timeTaken() // 1000)
 
     def _showAnswerButton(self):
         self._bottomReady = True
@@ -656,7 +695,8 @@ function showAnswer(txt) {
     def showContextMenu(self):
         opts = [
             [_("Mark Note"), "*", self.onMark],
-            [_("Bury Note"), "-", self.onBuryNote],
+            [_("Bury Card"), "-", self.onBuryCard],
+            [_("Bury Note"), "=", self.onBuryNote],
             [_("Suspend Card"), "@", self.onSuspendCard],
             [_("Suspend Note"), "!", self.onSuspend],
             [_("Delete Note"), "Delete", self.onDelete],
@@ -675,6 +715,7 @@ function showAnswer(txt) {
             a = m.addAction(label)
             a.setShortcut(QKeySequence(scut))
             a.connect(a, SIGNAL("triggered()"), func)
+        runHook("Reviewer.contextMenuEvent",self,m)
         m.exec_(QCursor.pos())
 
     def onOptions(self):
@@ -717,6 +758,12 @@ function showAnswer(txt) {
             "Note and its %d cards deleted.",
             cnt) % cnt)
 
+    def onBuryCard(self):
+        self.mw.checkpoint(_("Bury"))
+        self.mw.col.sched.buryCards([self.card.id])
+        self.mw.reset()
+        tooltip(_("Card buried."))
+
     def onBuryNote(self):
         self.mw.checkpoint(_("Bury"))
         self.mw.col.sched.buryNote(self.card.nid)
