@@ -3,11 +3,11 @@
 # License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
 
 import sys
+from anki.hooks import runHook
 from aqt.qt import *
 from aqt.utils import openLink
 from anki.utils import isMac, isWin
 import anki.js
-QtConfig = pyqtconfig.Configuration()
 
 # Bridge for Qt<->JS
 ##########################################################################
@@ -40,7 +40,7 @@ class AnkiWebPage(QWebPage):
 
 class AnkiWebView(QWebView):
 
-    def __init__(self):
+    def __init__(self, canFocus=True):
         QWebView.__init__(self)
         self.setRenderHints(
             QPainter.TextAntialiasing |
@@ -59,6 +59,7 @@ class AnkiWebView(QWebView):
         self.allowDrops = False
         # reset each time new html is set; used to detect if still in same state
         self.key = None
+        self.setCanFocus(canFocus)
 
     def keyPressEvent(self, evt):
         if evt.matches(QKeySequence.Copy):
@@ -78,14 +79,13 @@ class AnkiWebView(QWebView):
         QWebView.keyReleaseEvent(self, evt)
 
     def contextMenuEvent(self, evt):
-        # lazy: only run in reviewer
-        import aqt
-        if aqt.mw.state != "review":
+        if not self._canFocus:
             return
         m = QMenu(self)
         a = m.addAction(_("Copy"))
-        a.connect(a, SIGNAL("activated()"),
+        a.connect(a, SIGNAL("triggered()"),
                   lambda: self.triggerPageAction(QWebPage.Copy))
+        runHook("AnkiWebView.contextMenuEvent", self, m)
         m.popup(QCursor.pos())
 
     def dropEvent(self, evt):
@@ -130,6 +130,13 @@ button {
     def setBridge(self, bridge):
         self._bridge.setBridge(bridge)
 
+    def setCanFocus(self, canFocus=False):
+        self._canFocus = canFocus
+        if self._canFocus:
+            self.setFocusPolicy(Qt.WheelFocus)
+        else:
+            self.setFocusPolicy(Qt.NoFocus)
+
     def eval(self, js):
         self.page().mainFrame().evaluateJavaScript(js)
 
